""" pytest module to hold tests for webstaterator/__init__.py """

import os
import shutil

import pytest

from webstaterator.webstaterator import Webstaterator

from .test_website import get_basic_website
from .test_website import get_valid_pages
from .test_website import get_invalid_pages
from .test_website import get_valid_pages_with_lists
from .test_website import create_page
from .test_templateengine import make_template

#--- validator tests -----------------------------------------------------------
def create_path(tmpdir, name):
    new_path = os.path.join(str(tmpdir),name)

    if os.path.isdir(new_path):
        shutil.rmtree(new_path)

    os.mkdir(new_path)
    return new_path

def build_dummy_website(tmpdir, website, template_values = {}):
    path = os.path.join(tmpdir,website["template"])

    for page in website["pages"]:
        value = "test_template"
        if page["template"] in template_values:
            value = template_values[page["template"]]

        make_template(path, page["template"], value)

def assert_expected_files(build_path,expected_files):
    files = os.listdir(build_path)

    for file in expected_files:
        assert file in files


def test_validator_with_good_website(capsys, tmpdir):
    test_website = get_basic_website(get_valid_pages())
    test_website["template"] = str(tmpdir)
    build_dummy_website(tmpdir, test_website)

    weberator = Webstaterator()
    weberator.validate(test_website)
    captured = capsys.readouterr()
    assert captured.out.strip() == 'Ok'

def test_validator_with_good_website_with_assets(capsys, tmpdir):
    test_website = get_basic_website(get_valid_pages())
    test_website["template"] = str(tmpdir)
    test_website["assets"] = [str(tmpdir)]
    build_dummy_website(tmpdir, test_website)

    weberator = Webstaterator()
    weberator.validate(test_website)
    captured = capsys.readouterr()
    assert captured.out.strip() == 'Ok'

def test_validator_with_missing_templates(capsys):
    test_website = get_basic_website(get_valid_pages())

    weberator = Webstaterator()
    weberator.validate(test_website)
    captured = capsys.readouterr()
    assert captured.out.strip() != 'Ok'

def test_validator_with_missing_assets(capsys):
    test_website = get_basic_website(get_valid_pages())
    test_website["assets"] = "missing"

    weberator = Webstaterator()
    weberator.validate(test_website)
    captured = capsys.readouterr()
    assert captured.out.strip() != 'Ok'

def test_validator_with_bad_website_invalid_pages(capsys):

    weberator = Webstaterator()
    weberator.validate(get_basic_website(get_invalid_pages()))
    captured = capsys.readouterr()
    assert captured.out.strip() != 'Ok'

#--- build tests ---------------------------------------------------------------
def test_build_basic_website(capsys, tmpdir):
    build_path = create_path(tmpdir,"build")
    input_path =  create_path(tmpdir,"input")

    test_pages = get_valid_pages()

    test_website = get_basic_website(test_pages)
    test_website["template"] = input_path
    build_dummy_website(tmpdir, test_website)

    weberator = Webstaterator()

    weberator.build(test_website, build_path)

    for page in test_pages:
        assert os.path.exists(os.path.join(build_path,page["filename"]))

# test asset/
# test asset1 & 2
@pytest.mark.parametrize('assets',[
    (1, ['assets']),
    (2, ['assets_sep/']),
    (3, ['asset1','asset2'])]
)
def test_build_with_assets(tmpdir,assets):

    build_path =  create_path(tmpdir,f"build-{assets[0]}")
    input_path = create_path(tmpdir,"input")

    assets = assets[1]

    test_asset_format = "test-{}.css"

    asset_index = 1
    asset_paths = []
    for asset in assets:
        asset_paths.append(create_path(tmpdir,asset))

        asset_name = test_asset_format.format(asset_index)
        test_asset = os.path.join(asset_paths[-1],asset_name)
        with open(test_asset,"w") as asset_out:
            asset_out.write("BODY{background-color:#FF00FF;}")
        asset_index += 1

    test_website = get_basic_website(get_valid_pages())
    test_website["template"] = input_path
    test_website["assets"] = asset_paths
    build_dummy_website(tmpdir, test_website)

    weberator = Webstaterator()

    weberator.build(test_website, build_path)

    asset_index = 1
    for asset in assets:
        assert os.path.exists(os.path.join(build_path,asset)) == True
        assert os.path.exists(os.path.join(build_path,asset,test_asset_format.format(asset_index))) == True
        asset_index += 1


def test_build_model_access(tmpdir):
    build_path =  create_path(tmpdir,"build")
    input_path =  create_path(tmpdir,"input")
    item_template = """{% for item in model["values"] %}{{ item["name"] }}-{% endfor %}"""
    item_details = { "index.html": item_template }

    test_website = get_basic_website([create_page("index","index.html","index.html")])
    test_website["template"] = input_path
    model = {"values":[
        {"id":1, "name":"foo"},
        {"id":2, "name":"bar"},
        {"id":3, "name":"barfoo"}
    ]}
    test_website["model"] = model
    build_dummy_website(tmpdir, test_website, item_details)
    weberator = Webstaterator()
    weberator.build(test_website, build_path)

    assert_expected_files(build_path,["index.html"])

    test_page_path = os.path.join(build_path,"index.html")
    expected_contents = "-".join([obj["name"] for obj in model["values"]]) + "-"
    with open(test_page_path,'r') as page_reader:
        assert page_reader.read() == expected_contents

def test_build_page_lists_and_targets(tmpdir):
    build_path =  create_path(tmpdir,"build")
    input_path =  create_path(tmpdir,"input")

    item_template = """{{ target["name"] }}"""
    item_details = { "items.html": item_template }

    test_website = get_basic_website(get_valid_pages_with_lists())
    test_website["template"] = input_path
    models = [
        {"id":1, "name":"foo"},
        {"id":2, "name":"bar"},
        {"id":3, "name":"barfoo"}
    ]
    test_website["model"] = { "items": models }
    build_dummy_website(tmpdir, test_website, item_details)
    weberator = Webstaterator()
    weberator.build(test_website, build_path)

    expected_files = [
        'index.html',
        'items.html',
        'item-1.html',
        'item-2.html',
        'item-3.html'
    ]
    assert_expected_files(build_path,expected_files)

    for model in models:
        page_path = os.path.join(build_path,"item-{}.html".format(model['id']))
        with open(page_path,'r') as page_reader:
            assert page_reader.read() == model['name']

def test_build_links(tmpdir):
    build_path =  create_path(tmpdir,"build")
    input_path =  create_path(tmpdir,"input")
    item_template = """{% for name,urls in links.items() %}{{ name }}-{% endfor %}"""
    item_details = { "links.html": item_template }

    test_website = get_basic_website(get_valid_pages())
    test_website["template"] = input_path
    build_dummy_website(tmpdir, test_website, item_details)
    weberator = Webstaterator()
    weberator.build(test_website, build_path)

    expected_files = [
        'index.html',
        'items.html',
    ]

    assert_expected_files(build_path,expected_files)

    test_page_path = os.path.join(build_path,"items.html")
    expected_contents = "-".join([page["name"] for page in get_valid_pages()]) + "-"
    with open(test_page_path,'r') as page_reader:
        assert page_reader.read() == expected_contents



"""
Holds all the details for a website
"""

import os
import json
from collections.abc import Mapping

class InvalidPagesException(Exception):
    """ Exception used to indecate an invalid page object. Raised by Website """
    def __init__(self,msg,error = "Page model invalid"):
        self.error = error
        super().__init__(msg)

class InvalidModelException(Exception):
    """ Exception used to indecate an invalid data object. Raised by Website """
    def __init__(self,msg,error = "Data model invalid"):
        self.error = error
        super().__init__(msg)

class InvalidWebsiteException(Exception):
    """ Exception used to indecate an invalid website object. Raised by Website """
    def __init__(self,msg,error = "Website model invalid"):
        self.error = error
        super().__init__(msg)

class WebsiteAttributes:
    # pylint: disable=R0903
    # Too few public methods - Didn't find a nicer way to do this
    """ Holds string values for all website attributes """
    NAME = 'name'
    MODEL = 'model'
    TEMPLATE = 'template'
    ASSETS = 'assets'
    PAGES = 'pages'

    OPTIONAL_ATTRIBUTES = [
        ASSETS,
        MODEL
    ]

    REQUIRED_ATTRIBUTES = [
        NAME,
        TEMPLATE,
        PAGES
    ]

class PageAttributes:
    # pylint: disable=R0903
    # Too few public methods - Didn't find a nicer way to do this
    """ Holds string values for all page attributes """
    LIST = 'list'
    ID = 'id'
    NAME ='name'
    TEMPLATE = 'template'
    FILENAME = 'filename'

    OPTIONAL_ATTRIBUTES = [
        LIST,
        ID,
    ]
    REQUIRED_ATTRIBUTES = [
        NAME,
        TEMPLATE,
        FILENAME,
    ]

MODEL_ERR_FMT = "'{}' missing from model '{}'"
PAGE_ERR_FMT = "multipages must contain a '{{}}' in their {}"

class Website: # pylint: disable=R0902
    """ Represents a Website, holds detail on the pages to generated
        and the data model to use during generation
    """

    def __init__(self):
        """ Website represents the various aspects of the website """
        self._name = None
        self._template_path = None
        self._asset_paths = None
        self._links = {}
        self._targets = {}
        self._has_generated_links = False
        self._model = {}
        self._pages = []

#--- load ----------------------------------------------------------------------
    @staticmethod
    def __process_path(target_path):
        """ Checks target_path exists and handles relative and user paths """
        if '~' in target_path:
            target_path = os.path.expanduser(target_path)

        if target_path.endswith(os.path.sep):
            target_path = target_path[:-1]

        if not os.path.isdir(target_path):
            target_path = os.path.join(os.getcwd(),target_path)
            if not os.path.isdir(target_path):
                err_str = "Template folder {} not found".format(target_path)
                raise InvalidWebsiteException(err_str)
        return target_path

    def load(self,website):
        """ Load in a website either from json string, a dictionary, or a file path """
        self._has_generated_links = False
        website_obj = {}
        if isinstance(website, Mapping):
            website_obj = website
        else:
            website_obj = Website.__set_json_from_str_or_file(website)
            if os.path.isfile(website) and len(os.path.dirname(website)) > 0:
                #print(f"changing dir to {os.path.dirname(website)}" )
                os.chdir(os.path.dirname(website))

        Website.__validate_website(website_obj)

        self._template_path = Website.__process_path(
            website_obj[WebsiteAttributes.TEMPLATE]
        )

        self._name = website_obj[WebsiteAttributes.NAME]
        self._pages = website_obj[WebsiteAttributes.PAGES]

        if WebsiteAttributes.MODEL in website_obj:
            self.model = website_obj[WebsiteAttributes.MODEL]

        if WebsiteAttributes.ASSETS in website_obj:
            assets = website_obj[WebsiteAttributes.ASSETS]

            self._asset_paths = [
                Website.__process_path(asset)
                for asset in assets
            ]

    @staticmethod
    def __validate_website(website):
        """ Inspects website dictionary to check it has all required attributes """
        for required_attr in WebsiteAttributes.REQUIRED_ATTRIBUTES:
            if required_attr not in website:
                err_str = "website is missing {}".format(required_attr)
                raise InvalidWebsiteException(err_str)

        if Website.__validate_pages(website[WebsiteAttributes.PAGES]):
            return True
        return False

#--- basic properties ----------------------------------------------------------
    @property
    def name(self):
        """ get for name """
        return self._name

    @name.setter
    def name(self, value):
        """ set name """
        self._name = value

    @property
    def template_path(self):
        """ get for template_path """
        return self._template_path

    @template_path.setter
    def template_path(self, value):
        """ set template_path """
        self._template_path = value

    @property
    def asset_paths(self):
        """ get for asset paths """
        return self._asset_paths

    @asset_paths.setter
    def asset_paths(self, value):
        """ set for asset_paths
            :value: List of asset paths
        """
        self._asset_paths = value

#--- model ---------------------------------------------------------------------
    @property
    def model(self):
        """ get for model """
        return self._model

    @model.setter
    def model(self, value):
        """ set for model. Clears links
            :value: Either a dict, a json string, or a json file
        """
        self._has_generated_links = False
        if isinstance(value, Mapping):
            self._model = value
        else:
            self._model = Website.__set_json_from_str_or_file(value)

#--- pages ---------------------------------------------------------------------

    @property
    def pages(self):
        """ get pages """
        return self._pages

    @pages.setter
    def pages(self,value):
        """ set pages. Clears links
            :value: Either a dict, a json string or a json file
        """
        new_pages = []
        self._has_generated_links = False
        if isinstance(value,list): #we need to rethink this
            new_pages = value
        else:
            new_pages = Website.__set_json_from_str_or_file(value)

        if Website.__validate_pages(new_pages):
            self._pages = new_pages

    @staticmethod
    def __validate_pages(pages):
        """ Validate pages model
            :returns: True if valid
            :rases: InvalidPagesException if invalid
        """

        if len(pages) == 0:
            raise InvalidPagesException("No pages found!")

        if not isinstance(pages,list):
            err_str = "Expected list of pages, found {}".format(type(pages))
            raise InvalidPagesException(err_str)

        page_names = []

        for page in pages:
            for required_attr in PageAttributes.REQUIRED_ATTRIBUTES:
                if required_attr not in page:
                    page_name = "???"
                    if 'name' in page:
                        page_name = page['name']
                    err_str = "Required field '{}' is missing from the page called {}."
                    raise InvalidPagesException(err_str.format(required_attr,page_name))
            if page['name'] in page_names:
                err_str = "Page names must be unique. The name {} is used multiple times"
                raise InvalidPagesException(err_str.format(page['name']))

            page_names.append(page['name'])

        return True

#--- helpers -------------------------------------------------------------------
    @staticmethod
    def __set_json_from_str_or_file(value):
        obj = {}
        if os.path.isfile(value):
            with open(value,'r') as obj_in:
                obj = json.load(obj_in)
        else:
            obj = json.loads(value)
        return obj

    def to_dict(self, model_value = "model"):
        """ converts the website object to a dictionary """
        web_obj = {}
        web_obj[WebsiteAttributes.NAME] = self._name
        web_obj[WebsiteAttributes.MODEL] = model_value
        web_obj[WebsiteAttributes.TEMPLATE] = self._template_path
        web_obj[WebsiteAttributes.ASSETS] = self._asset_paths
        web_obj[WebsiteAttributes.PAGES] = self._pages

        return web_obj

#--- links ---------------------------------------------------------------------
    @staticmethod
    def validate_list_page(page):
        """ checks a list page object has all required elements"""
        if '{}' not in page[PageAttributes.FILENAME]:
            err_str = PAGE_ERR_FMT.format(PageAttributes.FILENAME)
            raise InvalidPagesException(err_str)
        return True

    @staticmethod
    def get_target_model(target_obj, obj):
        """ finds the target element within a dict """
        if '.' in target_obj:
            nestedmodel_obj = target_obj.split('.')
            breadcrumb = ""
            for obj_target in nestedmodel_obj:
                if obj_target not in obj:
                    err_str = MODEL_ERR_FMT.format(obj_target, breadcrumb[0:-1])
                    raise InvalidModelException(err_str)
                obj = obj[obj_target]
        else:
            if target_obj not in obj:
                err_str = MODEL_ERR_FMT.format(target_obj,'model')
                raise InvalidModelException(err_str)

            obj = obj[target_obj]
        return obj

    def generate_links(self):
        """ Generates based on a combination of model and pages """
        self._has_generated_links = False
        self._links.clear()
        self._targets.clear()

        err_str = ""

        for page in self._pages:
            if PageAttributes.LIST not in page:
                self._links[page[PageAttributes.NAME]] = [page[PageAttributes.FILENAME]]
                continue

            self.validate_list_page(page)
            model_obj = page[PageAttributes.LIST]

            obj = self.get_target_model(model_obj, self._model)

            item_links = []
            if isinstance(obj, Mapping):
                for key, item_obj in obj.items():
                    url = page[PageAttributes.FILENAME].format(key)
                    item_links.append(url)
                    self._targets[url] = item_obj
            else:
                if PageAttributes.ID not in page:
                    err_str = "Pages using the list attribute must include "
                    err_str+= "an id attribute or target a dictionary"
                    raise InvalidPagesException(err_str)
                model_id  = page[PageAttributes.ID]

                index = 0
                for item_obj in obj:
                    if model_id not in item_obj:
                        err_str = "{} at index {} is missing {}"
                        err_str = err_str.format(model_obj, index, model_id)
                        raise InvalidModelException(err_str)

                    url = page[PageAttributes.FILENAME].format(item_obj[model_id])
                    item_links.append(url)
                    self._targets[url] = item_obj
                    index += 1

            self._links[page[PageAttributes.NAME]] = item_links


        self._has_generated_links = len(self._links) > 0
        return self._has_generated_links

    @property
    def links(self):
        """ getter for _links """
        return self._links

    def get_target(self,link):
        """ returns the target for a given link or none """
        return self._targets.get(link, None)



"""
Webstaterator

A Python tool for generating static websites based on object models

Documentation: https://gitlab.com/Jon.Keatley.Folio/webstaterator
Gitlab: https://gitlab.com/Jon.Keatley.Folio/webstaterator
PyPi: https://pypi.org/project/webstaterator/

Created by Jon Keatley (http://jon-keatley.com)
Named by Sasha Siegel. It is her fault!

Copyright Jon Keatley 2021

"""

import sys

from webstaterator.cli import get_parser_options
from webstaterator.webstaterator import Webstaterator


def run(args):
    """ Execute webstaterators CLI """
    pargs = get_parser_options().parse_args(args)

    if pargs.action is None:
        get_parser_options().parse_args(["-h"])
        sys.exit(0)

    ws_obj = Webstaterator()

    if pargs.action == 'validate':
        ws_obj.validate(pargs.website)
    elif pargs.action == 'build':
        ws_obj.build(pargs.website, pargs.output)
    elif pargs.action == 'template':
        ws_obj.template(pargs.output)

def execute():
    """ wrapper method that calls run with sys.argv[1:] """
    run(sys.argv[1:])

if __name__ == '__main__':
    execute()



from setuptools import setup, find_packages

import os

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="webstaterator",
    version="0.3.1",
    packages=find_packages(),
    install_requires=["jinja2 >= '2.11.2'"],
    author='Jon Keatley',
    keywords='website static web pages generator',
    description='A Python tool for generating static websites based on object models',
    url='https://gitlab.com/Jon.Keatley.Folio/webstaterator',
    project_urls={'Source Code':'https://gitlab.com/Jon.Keatley.Folio/webstaterator'},
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3.5',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={'console_scripts':[
     'webstaterator = webstaterator.__main__:execute']},
)
